/****************************************************************************
 *
 * CRI Middleware SDK
 *
 * Copyright (c) 2017-2021 CRI Middleware Co., Ltd.
 *
 * Library  : CRIWARE plugin for Unreal Engine 4
 * Module   : Movie Scene Atom Template
 * File     : MovieSceneAtomTemplate.cpp
 *
 ****************************************************************************/

#include "MovieSceneAtomTemplate.h"

#include "TimerManager.h"
#include "Engine/EngineTypes.h"
#include "GameFramework/Actor.h"
#include "MovieScene.h"
#include "Evaluation/MovieSceneEvaluation.h"
#include "UObject/Package.h"
#include "GameFramework/WorldSettings.h"
#include "IMovieScenePlayer.h"

#include "MovieSceneAtomSection.h"
#include "MovieSceneAtomTrack.h"


DECLARE_CYCLE_STAT(TEXT("Atom Track Evaluate"), MovieSceneEval_AtomTrack_Evaluate, STATGROUP_MovieSceneEval);
DECLARE_CYCLE_STAT(TEXT("Atom Track Tear Down"), MovieSceneEval_AtomTrack_TearDown, STATGROUP_MovieSceneEval);
DECLARE_CYCLE_STAT(TEXT("Atom Track Token Execute"), MovieSceneEval_AtomTrack_TokenExecute, STATGROUP_MovieSceneEval);

static float MaxSequenceAtomDesyncToleranceCVar = 0.5f;
FAutoConsoleVariableRef CVarMaxSequenceAtomDesyncTolerance(
	TEXT("Sequencer.Atom.MaxDesyncTolerance"),
	MaxSequenceAtomDesyncToleranceCVar,
	TEXT("Controls how many seconds an atom track can be out of sync in a Sequence before we attempt a time correction.\n"),
	ECVF_Default);

static bool bIgnoreAtomSyncDuringWorldTimeDilationCVar = true;
FAutoConsoleVariableRef CVarIgnoreAtomSyncDuringWorldTimeDilation(
	TEXT("Sequencer.Atom.IgnoreAtomSyncDuringWorldTimeDilation"),
	bIgnoreAtomSyncDuringWorldTimeDilationCVar,
	TEXT("Ignore correcting audio if there is world time dilation.\n"),
	ECVF_Default);

static int32 UseAtomClockForSequencerDesyncCVar = 0;
FAutoConsoleVariableRef CVaUseAtomClockForSequencerDesync(
	TEXT("Sequencer.Atom.UseAtomClockForAudioDesync"),
	UseAtomClockForSequencerDesyncCVar,
	TEXT("When set to 1, we will use the atom render thread directly to query whether audio has went out of sync with the sequence.\n"),
	ECVF_Default);

/** Stop audio from playing */
struct FStopAtomPreAnimatedToken : IMovieScenePreAnimatedToken
{
	static FMovieSceneAnimTypeID GetAnimTypeID()
	{
		return TMovieSceneAnimTypeID<FStopAtomPreAnimatedToken>();
	}

#if ENGINE_MAJOR_VERSION == 4 && ENGINE_MINOR_VERSION <= 26
	virtual void RestoreState(UObject& InObject, IMovieScenePlayer& Player) override
#else
	virtual void RestoreState(UObject& InObject, const UE::MovieScene::FRestoreStateParams& Params) override
#endif
	{
		UAtomComponent* AtomComponent = CastChecked<UAtomComponent>(&InObject);
		bool bAtomComponentStatus = (AtomComponent->GetStatus() == EAtomComponentStatus::Prep) || (AtomComponent->GetStatus() == EAtomComponentStatus::Playing);
		if (AtomComponent && (AtomComponent->bAutoDestroy || !bAtomComponentStatus)) {
			/* AtomComponentの再生継続設定が有効でない場合はAtomComponentを破棄する */
			AtomComponent->Stop();
			AtomComponent->DestroyComponent();
		}
	}

	struct FProducer : IMovieScenePreAnimatedTokenProducer
	{
		virtual IMovieScenePreAnimatedTokenPtr CacheExistingState(UObject& Object) const override
		{
			return FStopAtomPreAnimatedToken();
		}
	};
};

/** Destroy a transient audio component */
struct FDestroyAtomPreAnimatedToken : IMovieScenePreAnimatedToken
{
	static FMovieSceneAnimTypeID GetAnimTypeID()
	{
		return TMovieSceneAnimTypeID<FDestroyAtomPreAnimatedToken>();
	}

#if ENGINE_MAJOR_VERSION == 4 && ENGINE_MINOR_VERSION <= 26
	virtual void RestoreState(UObject& InObject, IMovieScenePlayer& Player) override
#else
	virtual void RestoreState(UObject& InObject, const UE::MovieScene::FRestoreStateParams& Params) override
#endif
	{
		UAtomComponent* AtomComponent = CastChecked<UAtomComponent>(&InObject);
		bool bAtomComponentStatus = (AtomComponent->GetStatus() == EAtomComponentStatus::Prep) || (AtomComponent->GetStatus() == EAtomComponentStatus::Playing);
		if (AtomComponent && (AtomComponent->bAutoDestroy || !bAtomComponentStatus)) {
			/* シーケンサーの破棄時にAtomComponentの再生継続設定が有効になっていない場合はそのまま破棄する */
			AtomComponent->DestroyComponent();
		}
		else
		{
			/* AtomComponentの再生継続設定が有効になっている場合停止したらAtomComponentを破棄するようbAutoDestroyを有効にしておく */
			AtomComponent->bAutoDestroy = true;
		}
	}

	struct FProducer : IMovieScenePreAnimatedTokenProducer
	{
		virtual IMovieScenePreAnimatedTokenPtr CacheExistingState(UObject& Object) const override
		{
			return FDestroyAtomPreAnimatedToken();
		}
	};
};

struct FCachedAtomTrackData : IPersistentEvaluationData
{
	TMap<FObjectKey, TMap<FObjectKey, TWeakObjectPtr<UAtomComponent>>> AtomComponentsByActorKey;

	FCachedAtomTrackData()
	{
		/* マスタートラックのコンテナ作成 */
		AtomComponentsByActorKey.Add(FObjectKey(), TMap<FObjectKey, TWeakObjectPtr<UAtomComponent>>());
	}

	/* AtomComponent に対して再生要求が発行されるたびに、この TMap に要素をセットします。 */
	/* この TMap は、シーケンサが実時間よりも高速に評価された場合の同期ズレを検出するのに使用します。 */
	TMap<TWeakObjectPtr<UAtomComponent>, float> SoundLastPlayedAtTime;

	UAtomComponent* GetAtomComponent(FObjectKey ActorKey, FObjectKey SectionKey)
	{
		if (TMap<FObjectKey, TWeakObjectPtr<UAtomComponent>>* Map = AtomComponentsByActorKey.Find(ActorKey))
		{
			/* まず始めにセクションにマッチするAtomComponentがすでにないか確認 */
			TWeakObjectPtr<UAtomComponent> ExistingComponentPtr = Map->FindRef(SectionKey);
			if (ExistingComponentPtr.IsValid())
			{
				return ExistingComponentPtr.Get();
			}

			/* AtomComponentが見つからなかった場合は、Map中に使われていないAtomComponentがないか確認する */
			for (TPair<FObjectKey, TWeakObjectPtr<UAtomComponent >> Pair : *Map)
			{
				UAtomComponent* ExistingComponent = Map->FindRef(Pair.Key).Get();
				if (ExistingComponent && !ExistingComponent->IsPlaying())
				{
					/* 新しいセクションのAtomComponentとして登録する */ 
					Map->Remove(Pair.Key);
					Map->Add(SectionKey, ExistingComponent);
					return ExistingComponent;
				}
			}
		}

		return nullptr;
	}

	/* サブトラック用のAtomComponentを生成 */
	UAtomComponent* AddAtomComponentForRow(int32 RowIndex, FObjectKey SectionKey, UObject& PrincipalObject, IMovieScenePlayer& Player)
	{
		FObjectKey ObjectKey(&PrincipalObject);

		if (!AtomComponentsByActorKey.Contains(ObjectKey))
		{
			AtomComponentsByActorKey.Add(ObjectKey, TMap<FObjectKey, TWeakObjectPtr<UAtomComponent>>());
		}

		UAtomComponent* ExistingComponent = GetAtomComponent(ObjectKey, SectionKey);
		if (!ExistingComponent)
		{
			/* AtomComponentが存在しない場合は生成する */

			USoundAtomCue* TempPlaybackAudioCue = NewObject<USoundAtomCue>();

			AActor* Actor = nullptr;
			USceneComponent* SceneComponent = nullptr;
			FString ObjectName;

			if (PrincipalObject.IsA<AActor>())
			{
				Actor = Cast<AActor>(&PrincipalObject);
				SceneComponent = Actor->GetRootComponent();
				ObjectName =
#if WITH_EDITOR
					Actor->GetActorLabel();
#else
					Actor->GetName();
#endif
			}
			else if (PrincipalObject.IsA<UActorComponent>())
			{
				UActorComponent* ActorComponent = Cast<UActorComponent>(&PrincipalObject);
				Actor = ActorComponent->GetOwner();
				SceneComponent = Cast<USceneComponent>(ActorComponent);
				ObjectName = ActorComponent->GetName();
			}

			if (!Actor || !SceneComponent)
			{
				UE_LOG(LogMovieScene, Warning, TEXT("Failed to find scene component for spatialized audio track (row %d)."), RowIndex);
				return nullptr;
			}


			UObject* PlaybackContext = Player.GetPlaybackContext();
			UWorld* World = PlaybackContext ? PlaybackContext->GetWorld() : nullptr;
			AActor* MyActor = (World ? World->GetWorldSettings() : nullptr);
			if (MyActor != nullptr) {
				ExistingComponent = NewObject<UAtomComponent>(MyActor);
			}
			else {
				ExistingComponent = NewObject<UAtomComponent>();
			}

			if (!ExistingComponent)
			{
				UE_LOG(LogMovieScene, Warning, TEXT("Failed to create audio component for spatialized audio track (row %d on %s)."), RowIndex, *ObjectName);
				return nullptr;
			}

#if WITH_EDITOR
			bool bIsPreview = GIsEditor && World != nullptr && !World->IsPlayInEditor();
			if (!bIsPreview)
			{
#endif
				/* 作成したAtomComponentをRegisterComponentをすることでTick動作するようにする */
				ExistingComponent->RegisterComponent();
#if WITH_EDITOR
			}
#endif

			Player.SavePreAnimatedState(*ExistingComponent, FMovieSceneAnimTypeID::Unique(), FDestroyAtomPreAnimatedToken::FProducer());

			AtomComponentsByActorKey[ObjectKey].Add(SectionKey, ExistingComponent);

			ExistingComponent->SetFlags(RF_Transient);
			ExistingComponent->AttachToComponent(SceneComponent, FAttachmentTransformRules::KeepRelativeTransform);
		}

		return ExistingComponent;
	}

	/* マスタートラック用のAtomComponentを生成 */
	UAtomComponent* AddMasterAtomComponentForRow(int32 RowIndex, FObjectKey SectionKey, UWorld* World, IMovieScenePlayer& Player)
	{
		UAtomComponent* ExistingComponent = GetAtomComponent(FObjectKey(), SectionKey);
		if (!ExistingComponent)
		{
			USoundAtomCue* TempPlaybackAtomCue = NewObject<USoundAtomCue>();

			AActor* Actor = (World ? World->GetWorldSettings() : nullptr);
			if (Actor != nullptr) {
				ExistingComponent = NewObject<UAtomComponent>(Actor);
			} else {
				ExistingComponent = NewObject<UAtomComponent>();
			}

			if (!ExistingComponent)
			{
				UE_LOG(LogMovieScene, Warning, TEXT("Failed to create audio component for master audio track (row %d)."), RowIndex);
				return nullptr;
			}

#if WITH_EDITOR
			bool bIsPreview = GIsEditor && World != nullptr && !World->IsPlayInEditor();
			if (!bIsPreview)
			{
#endif
				/* 作成したAtomComponentをRegisterComponentをすることでTick動作するようにする */
				ExistingComponent->RegisterComponent();
#if WITH_EDITOR
			}
#endif

			Player.SavePreAnimatedState(*ExistingComponent, FMovieSceneAnimTypeID::Unique(), FDestroyAtomPreAnimatedToken::FProducer());

			ExistingComponent->SetFlags(RF_Transient);

			AtomComponentsByActorKey[FObjectKey()].Add(SectionKey, ExistingComponent);
		}

		return ExistingComponent;
	}

	void StopAllSounds()
	{
		for (TPair<FObjectKey, TMap<FObjectKey, TWeakObjectPtr<UAtomComponent>>>& Map : AtomComponentsByActorKey)
		{
			for (TPair<FObjectKey, TWeakObjectPtr<UAtomComponent>>& Pair : Map.Value)
			{
				if (UAtomComponent* AtomComponent = Pair.Value.Get())
				{
					AtomComponent->Stop();
				}
			}
		}
	}

	void StopSoundsOnSection(FObjectKey ObjectKey)
	{
		for (TPair<FObjectKey, TMap<FObjectKey, TWeakObjectPtr<UAtomComponent>>>& Pair : AtomComponentsByActorKey)
		{
			if (UAtomComponent* AtomComponent = Pair.Value.FindRef(ObjectKey).Get())
			{
				if (AtomComponent->bAutoDestroy) 
				{
					AtomComponent->Stop();
				}
			}
		}
	}
};

struct FAtomSectionExecutionToken : IMovieSceneExecutionToken
{
	FAtomSectionExecutionToken(const UMovieSceneAtomSection* InAudioSection)
		: AtomSection(InAudioSection), SectionKey(InAudioSection)
	{}

	virtual void Execute(const FMovieSceneContext& Context, const FMovieSceneEvaluationOperand& Operand, FPersistentEvaluationData& PersistentData, IMovieScenePlayer& Player)
	{
		FCachedAtomTrackData& TrackData = PersistentData.GetOrAddTrackData<FCachedAtomTrackData>();

		if ((Context.GetStatus() != EMovieScenePlayerStatus::Playing && Context.GetStatus() != EMovieScenePlayerStatus::Scrubbing) || Context.HasJumped() || Context.GetDirection() == EPlayDirection::Backwards) {
			// stopped, recording, etc
			TrackData.StopAllSounds();
		} else if (!Operand.ObjectBindingID.IsValid()) {
			/* マスタートラックでの処理 */

			UObject* PlaybackContext = Player.GetPlaybackContext();

			/* AtomTrackをアタッチするオブジェクトを取得 */
			const FMovieSceneActorReferenceData& AttachActorData = AtomSection->GetAttachActorData();

#if (ENGINE_MAJOR_VERSION == 4 && ENGINE_MINOR_VERSION >= 26) || ENGINE_MAJOR_VERSION == 5
			USceneComponent* AttachComponent = nullptr;
			FMovieSceneActorReferenceKey AttachKey;
			AttachActorData.Evaluate(Context.GetTime(), AttachKey);
			FMovieSceneObjectBindingID AttachBindingID = AttachKey.Object;

			if (AttachBindingID.IsValid())
			{
				FMovieSceneSequenceID SequenceID = Operand.SequenceID;
#if ENGINE_MAJOR_VERSION == 4 && ENGINE_MINOR_VERSION <= 26
				if (AttachBindingID.GetSequenceID().IsValid())
				{
					/* この ID が現在のローカルシーケンス ID に基づいてルートから解決可能であることを確認。 */
					FMovieSceneObjectBindingID RootBindingID = AttachBindingID.ResolveLocalToRoot(SequenceID, Player);
					SequenceID = RootBindingID.GetSequenceID();
				}
#else
				if (AttachBindingID.GetRelativeSequenceID().IsValid())
				{
					/* この ID が現在のローカルシーケンス ID に基づいてルートから解決可能であることを確認。 */
					FMovieSceneObjectBindingID RootBindingID = AttachBindingID.ResolveToFixed(SequenceID, Player);
					SequenceID = RootBindingID.GetRelativeSequenceID();
				}
#endif

				/* トランスフォームが設定されている場合、バインドされたアクタのトランスフォームを使用します。*/
				FMovieSceneEvaluationOperand ObjectOperand(SequenceID, AttachBindingID.GetGuid());

				for (TWeakObjectPtr<> WeakObject : Player.FindBoundObjects(ObjectOperand))
				{
					AActor* AttachActor = Cast<AActor>(WeakObject.Get());
					if (AttachActor)
					{
						AttachComponent = AtomSection->GetAttachComponent(AttachActor, AttachKey);
					}
					if (AttachComponent)
					{
						break;
					}
				}
			}

#endif
			UAtomComponent* AtomComponent = TrackData.GetAtomComponent(FObjectKey(), SectionKey);
			if (!AtomComponent) {
				// Initialize the sound
				AtomComponent = TrackData.AddMasterAtomComponentForRow(AtomSection->GetRowIndex(), SectionKey, PlaybackContext ? PlaybackContext->GetWorld() : nullptr, Player);
			}

			if (AtomComponent) {
#if (ENGINE_MAJOR_VERSION == 4 && ENGINE_MINOR_VERSION >= 26) || ENGINE_MAJOR_VERSION == 5
				/* AtomCompoentに取得したオブジェクトをアタッチする */
				if (AttachComponent && (AtomComponent->GetAttachParent() != AttachComponent || AtomComponent->GetAttachSocketName() != AttachKey.SocketName))
				{
					AtomComponent->AttachToComponent(AttachComponent, FAttachmentTransformRules::KeepRelativeTransform, AttachKey.SocketName);
				}
#endif
				EnsureAtomIsPlaying(*AtomComponent, PersistentData, Context, AtomComponent->GetAttachParent() != nullptr, Player);
			}
		} else {
			/* オブジェクトのほうがAtomTrackをバインディングしている場合 */
			for (TWeakObjectPtr<> Object : Player.FindBoundObjects(Operand)) {
				AActor* Actor = Cast<AActor>(Object.Get());
				if (!Actor) {
					continue;
				}

				UAtomComponent* AtomComponent = TrackData.GetAtomComponent(FObjectKey(), SectionKey);
				if (!AtomComponent) {
					/* AtomComponentがない場合は作成する */
					AtomComponent = TrackData.AddAtomComponentForRow(AtomSection->GetRowIndex(), SectionKey, *Object.Get(), Player);
				}

				if (AtomComponent) {
					EnsureAtomIsPlaying(*AtomComponent, PersistentData, Context, true, Player);
				}
			}
		}
	}

	void EnsureAtomIsPlaying(UAtomComponent& AtomComponent, FPersistentEvaluationData& PersistentData, const FMovieSceneContext& Context, bool bAllowSpatialization, IMovieScenePlayer& Player) const
	{
		Player.SavePreAnimatedState(AtomComponent, FStopAtomPreAnimatedToken::GetAnimTypeID(), FStopAtomPreAnimatedToken::FProducer());
		const_cast<UMovieSceneAtomSection*>(AtomSection)->SetAtomComponent(&AtomComponent);

		float AudioVolume = 1.f;
		AtomSection->GetSoundVolumeChannel().Evaluate(Context.GetTime(), AudioVolume);
		AudioVolume = AudioVolume * AtomSection->EvaluateEasing(Context.GetTime());
		AtomComponent.SetVolume(AudioVolume);

		float PitchMultiplier = 1.f;
		AtomSection->GetPitchMultiplierChannel().Evaluate(Context.GetTime(), PitchMultiplier);
		AtomComponent.SetPitchMultiplier(PitchMultiplier);

		TArray<FScalarParameterNameAndCurve> ScalarParameterNamesAndCurves = AtomSection->GetScalarParameterNamesAndCurves();
		for (auto ScalarParameterNamesAndCurve : ScalarParameterNamesAndCurves)
		{
			float AisacValue = 0.0f;
			ScalarParameterNamesAndCurve.ParameterCurve.Evaluate(Context.GetTime(), AisacValue);
			AtomComponent.SetAisacByName(ScalarParameterNamesAndCurve.ParameterName.ToString(), AisacValue);
		}

		/* ループパラメータに応じて、UAtomComponent のループ設定を変更する */
		if (AtomSection->IsLooping())
		{
			AtomComponent.LoopSetting = EAtomLoopSetting::Loop;
		}
		else
		{
			AtomComponent.LoopSetting = EAtomLoopSetting::OneShot;
		}

		AtomComponent.bAutoDestroy = !AtomSection->GetContinueSoundWhenSequenceIsEnd();
#if WITH_EDITOR
		UObject* PlaybackContext = Player.GetPlaybackContext();
		UWorld* World = PlaybackContext ? PlaybackContext->GetWorld() : nullptr;
		bool bIsPreview = GIsEditor && World != nullptr && !World->IsPlayInEditor();
		if (bIsPreview)
		{
			AtomComponent.bAutoDestroy = true;
		}
#endif

		float SectionStartTimeSeconds = (AtomSection->HasStartFrame() ? AtomSection->GetInclusiveStartFrame() : 0) / AtomSection->GetTypedOuter<UMovieScene>()->GetTickResolution();

		FCachedAtomTrackData& TrackData = PersistentData.GetOrAddTrackData<FCachedAtomTrackData>();
		const FFrameNumber AudioStartOffset = AtomSection->GetStartOffset();
		USoundAtomCue* Sound = AtomSection->GetSound();

		float AtomTime = (Context.GetTime() / Context.GetFrameRate()) - SectionStartTimeSeconds + (float)Context.GetFrameRate().AsSeconds(AudioStartOffset);
		if (AtomTime >= 0.f && Sound)
		{
			const float Duration = Sound->GetDuration();

			if (!AtomComponent.IsLoop() && AtomTime > Duration && Duration != 0.f)
			{
				AtomComponent.Stop();
				return;
			}

			AtomTime = Duration > 0.f ? FMath::Fmod(AtomTime, Duration) : AtomTime;
		}

		/* AtomComponentが再生されていない場合は、状態を変更する必要があります。AtomComponentが間違った音を再生している場合は、
		   状態を変更する必要があります。オーディオの再生時間が希望の時間と著しくずれている場合は、状態の変更が必要です。 */
		bool bSoundNeedsStateChange = !AtomComponent.IsPlaying() || AtomComponent.Sound != Sound;
		bool bSoundNeedsTimeSync = false;

		/* スローモーション等の時間拡張がない場合にのみ同期。これをしないとオーディオ再生が拡張されず、期待される再生時間と一致しないことになる。*/
		const bool bDoTimeSync =
			Player.GetPlaybackContext()->GetWorld() &&
			(FMath::IsNearlyEqual(Player.GetPlaybackContext()->GetWorld()->GetWorldSettings()->GetEffectiveTimeDilation(), 1.f) ||
				!bIgnoreAtomSyncDuringWorldTimeDilationCVar);

		if (bDoTimeSync)
		{
			float CurrentGameTime = 0.0f;

			CurrentGameTime = Player.GetPlaybackContext()->GetWorld()->GetAudioTimeSeconds();
			
			/* ゲーム中の時間とAudioの時間を比較してずれを計算。閾値時間外であった場合に時間のずれに関するログを出力する。 */
			if (TrackData.SoundLastPlayedAtTime.Contains(&AtomComponent))
			{
				float SoundLastPlayedAtTime = TrackData.SoundLastPlayedAtTime[&AtomComponent];

				float GameTimeDelta = CurrentGameTime - SoundLastPlayedAtTime;
				if (!FMath::IsNearlyZero(MaxSequenceAtomDesyncToleranceCVar) && FMath::Abs(GameTimeDelta - AtomTime) > MaxSequenceAtomDesyncToleranceCVar)
				{
					UE_LOG(LogMovieScene, Verbose, TEXT("Atom Component detected a significant mismatch in (assumed) playback time versus the desired time. Time since last play call: %6.2f(s) Desired Time: %6.2f(s). Component: %s sound: %s"), GameTimeDelta, AtomTime, *AtomComponent.GetName(), *GetNameSafe(AtomComponent.Sound));
					bSoundNeedsTimeSync = true;
				}
			}
		}

		if (bSoundNeedsStateChange || bSoundNeedsTimeSync)
		{
			if (AtomSection->GetOverrideAttenuation())
			{
				AtomComponent.AttenuationSettings = AtomSection->GetAttenuationSettings();
			}

			for (auto SelectorLabel : AtomSection->GetSelectorLabels())
			{
				AtomComponent.SetSelectorLabel(SelectorLabel.Selector, SelectorLabel.Label);
			}

			if (AtomSection->GetASRRackIDs().Num() > 0) {
				AtomComponent.SetAsrRackIdArray(AtomSection->GetASRRackIDs());
			}

			/* 実際に再生されている場合のみ、サウンドの停止を呼び出します。これにより、生成時間がゼロのサウンドキューが再生されている場合に、ストップ命令が連続して呼び出されることを防げます。 */
			if (AtomComponent.IsPlaying() || bSoundNeedsTimeSync)
			{
				UE_LOG(LogMovieScene, Verbose, TEXT("Atom Component stopped due to needing a state change bIsPlaying: %d bNeedsTimeSync: %d. Component: %s sound: %s"), AtomComponent.IsPlaying(), bSoundNeedsTimeSync, *AtomComponent.GetName(), *AtomComponent.Sound->GetName());
				AtomComponent.Stop();
			}

			/* サウンドクリップが実際に変更された場合のみ変更します。必要に応じて内部的にStopを呼び出します。 */
			if (AtomComponent.Sound != Sound)
			{
				UE_LOG(LogMovieScene, Verbose, TEXT("Atom Component calling SetSound due to new sound. Component: %s OldSound: %s NewSound: %s"), *AtomComponent.GetName(), *GetNameSafe(AtomComponent.Sound), *AtomComponent.Sound->GetName());
				AtomComponent.SetSound(Sound);
			}
#if WITH_EDITOR
			if (GIsEditor && World != nullptr && !World->IsPlayInEditor())
			{
				AtomComponent.bIsUISound = true;
				AtomComponent.bIsPreviewSound = true;
			}
			else
#endif // WITH_EDITOR
			{
				AtomComponent.bIsUISound = false;
			}

			if (AtomTime >= 0.f)
			{
				UE_LOG(LogMovieScene, Verbose, TEXT("Atom Component Play at Local Time: %6.2f CurrentTime: %6.2f(s) SectionStart: %6.2f(s), SoundDur: %6.2f OffsetIntoClip: %6.2f sound: %s"), AtomTime, (Context.GetTime() / Context.GetFrameRate()), SectionStartTimeSeconds, AtomComponent.Sound->GetDuration(), (float)Context.GetFrameRate().AsSeconds(AudioStartOffset), *AtomComponent.Sound->GetName());
				AsyncTask(ENamedThreads::AnyBackgroundThreadNormalTask, [&AtomComponent, AtomTime]() {
					if (AtomComponent.Sound && AtomComponent.Sound->CueSheet) 
					{
						AtomComponent.Sound->CueSheet->AsyncLoadCueSheet();
						while (!AtomComponent.Sound->CueSheet->IsLoaded())
						{
							FPlatformProcess::Sleep(0.0f);
						}
						AsyncTask(ENamedThreads::GameThread, [&AtomComponent, AtomTime]() {
							AtomComponent.Play(AtomTime);
						});
					}
				});

				// 後で再度再生が行われた際に、適切な時間で再生が行われるよう、AtomComponentとその時の再生時間を保存しておく.
				if (Player.GetPlaybackContext()->GetWorld())
				{
					if (!TrackData.SoundLastPlayedAtTime.Contains(&AtomComponent))
					{
						TrackData.SoundLastPlayedAtTime.Add(&AtomComponent);
					}

					TrackData.SoundLastPlayedAtTime[&AtomComponent] = Player.GetPlaybackContext()->GetWorld()->GetAudioTimeSeconds() - AtomTime;
				}
			}

			if (Context.GetStatus() == EMovieScenePlayerStatus::Scrubbing)
			{
#if WITH_EDITOR
				if (GIsEditor && World != nullptr && !World->IsPlayInEditor()) {
					/* エディタプレビュー時はカーソル付近の波形を一瞬だけ再生する */
					FTimerHandle PreviewStopTimerHandle;
					auto PreviewStopDelegate = FTimerDelegate::CreateLambda([&AtomComponent]() {
						AtomComponent.Stop();
						});
					World->GetTimerManager().SetTimer(PreviewStopTimerHandle,
						PreviewStopDelegate, AtomTrackConstants::ScrubDuration, false);
				}
#else
				AtomComponent.Stop();
#endif
			}
		}
	}

	const UMovieSceneAtomSection* AtomSection;
	FObjectKey SectionKey;
};

FMovieSceneAtomSectionTemplate::FMovieSceneAtomSectionTemplate(const UMovieSceneAtomSection& Section)
	: AtomSection(&Section)
{
}

void FMovieSceneAtomSectionTemplate::Evaluate(const FMovieSceneEvaluationOperand& Operand, const FMovieSceneContext& Context, const FPersistentEvaluationData& PersistentData, FMovieSceneExecutionTokens& ExecutionTokens) const
{
	MOVIESCENE_DETAILED_SCOPE_CYCLE_COUNTER(MovieSceneEval_AtomTrack_Evaluate)

#if WITH_EDITOR
	if (GEngine && GEngine->UseSound() && Context.GetStatus() != EMovieScenePlayerStatus::Jumping) {
#else
	if (GEngine && Context.GetStatus() != EMovieScenePlayerStatus::Jumping) {
#endif
		ExecutionTokens.Add(FAtomSectionExecutionToken(AtomSection));
	}
}

void FMovieSceneAtomSectionTemplate::TearDown(FPersistentEvaluationData & PersistentData, IMovieScenePlayer & Player) const
{
	MOVIESCENE_DETAILED_SCOPE_CYCLE_COUNTER(MovieSceneEval_AtomTrack_TearDown)
#if WITH_EDITOR
	if (GEngine && GEngine->UseSound())
#else
	if (GEngine)
#endif
	{
		FCachedAtomTrackData& TrackData = PersistentData.GetOrAddTrackData<FCachedAtomTrackData>();

		TrackData.StopSoundsOnSection(AtomSection);
	}
}
