/****************************************************************************
 *
 * CRI Middleware SDK
 *
 * Copyright (c) 2017-2021 CRI Middleware Co., Ltd.
 *
 * Library  : CRIWARE plugin for Unreal Engine 4
 * Module   : Atom Track Editor
 * File     : AtomTrackEditor.cpp
 *
 ****************************************************************************/

#include "AtomTrackEditor.h"

#include "Framework/Application/SlateApplication.h"
#include "Widgets/SBoxPanel.h"
#include "Widgets/Layout/SBox.h"
#include "Framework/MultiBox/MultiBoxBuilder.h"
#include "EditorStyleSet.h"
#include "SequencerUtilities.h"
#if ENGINE_MAJOR_VERSION == 5 && ENGINE_MINOR_VERSION > 0
#include "AssetRegistry/AssetRegistryModule.h"
#else
#include "AssetRegistryModule.h"
#endif
#include "IContentBrowserSingleton.h"
#include "ContentBrowserModule.h"
#include "ISectionLayoutBuilder.h"
#include "CommonMovieSceneTools.h"
#include "DragAndDrop/AssetDragDropOp.h"
#include "Misc/QualifiedFrameTime.h"

#include "MovieSceneAtomTrack.h"
#include "MovieSceneAtomSection.h"

#include "CriWareSequencerPrivate.h"
#include "Shared/CriWareSequencerEditorStyle.h"

#define LOCTEXT_NAMESPACE "FAtomTrackEditor"


FAtomTrackEditor::FAtomTrackEditor(TSharedRef<ISequencer> InSequencer)
	: FMovieSceneTrackEditor(InSequencer)
{
}

FAtomTrackEditor::~FAtomTrackEditor()
{
}

TSharedRef<ISequencerTrackEditor> FAtomTrackEditor::CreateTrackEditor(TSharedRef<ISequencer> InSequencer)
{
	return MakeShareable(new FAtomTrackEditor(InSequencer));
}

void FAtomTrackEditor::BuildAddTrackMenu(FMenuBuilder& MenuBuilder)
{
	MenuBuilder.AddMenuEntry(
		LOCTEXT("AddTrack", "Atom Track"),
		LOCTEXT("AddTooltip", "Adds a new master audio track that can play sounds."),
		//FSlateIcon(FEditorStyle::GetStyleSetName(), "Sequencer.Tracks.Atom"),
		FSlateIcon(FCriWareSequencerEditorStyle::Get().GetStyleSetName(), "ClassIcon.AtomPlane"),
		FUIAction(
			FExecuteAction::CreateRaw(this, &FAtomTrackEditor::HandleAddAtomTrackMenuEntryExecute)
		)
	);
}

void FAtomTrackEditor::BuildObjectBindingTrackMenu(FMenuBuilder& MenuBuilder, const TArray<FGuid>& ObjectBindings, const UClass* ObjectClass)
{
	if (ObjectClass != nullptr && ObjectClass->IsChildOf(AActor::StaticClass()))
	{
		/* ActorなどのオブジェクトにAtomTrackをサブトラックとして追加する処理 */
		MenuBuilder.AddSubMenu(
			LOCTEXT("AddAttachedAtomTrack", "Atom"),
			LOCTEXT("AddAttachedAtomTooltip", "Adds an atom track attached to the object."),
			FNewMenuDelegate::CreateSP(this, &FAtomTrackEditor::HandleAddAttachedAtomTrackMenuEntryExecute, ObjectBindings));
	}
}

bool FAtomTrackEditor::SupportsType(TSubclassOf<UMovieSceneTrack> Type) const
{
	return Type == UMovieSceneAtomTrack::StaticClass();
}

void FAtomTrackEditor::BuildTrackContextMenu(FMenuBuilder& MenuBuilder, UMovieSceneTrack* Track)
{
}

const FSlateBrush* FAtomTrackEditor::GetIconBrush() const
{
	//return FEditorStyle::GetBrush("Sequencer.Tracks.Audio");
	return FCriWareSequencerEditorStyle::Get().GetBrush("ClassIcon.AtomPlane");
}

bool FAtomTrackEditor::IsResizable(UMovieSceneTrack* InTrack) const
{
	return true;
}

void FAtomTrackEditor::Resize(float NewSize, UMovieSceneTrack* InTrack)
{
	UMovieSceneAtomTrack* AtomTrack = Cast<UMovieSceneAtomTrack>(InTrack);
	if (AtomTrack) {
		AtomTrack->Modify();

		int32 MaxNumRows = 1;
		for (UMovieSceneSection* Section : AtomTrack->GetAllSections()) {
			MaxNumRows = FMath::Max(MaxNumRows, Section->GetRowIndex() + 1);
		}

		AtomTrack->SetRowHeight(FMath::RoundToInt(NewSize) / MaxNumRows);
	}
}

#if ENGINE_MAJOR_VERSION == 4
bool FAtomTrackEditor::OnAllowDrop(const FDragDropEvent& DragDropEvent, UMovieSceneTrack* Track, int32 RowIndex, const FGuid& TargetObjectGuid)
#else
bool FAtomTrackEditor::OnAllowDrop(const FDragDropEvent& DragDropEvent, FSequencerDragDropParams& DragDropParams)
#endif
{
#if ENGINE_MAJOR_VERSION == 5
	if (!DragDropParams.Track.IsValid() || !DragDropParams.Track->IsA(UMovieSceneAtomTrack::StaticClass())) {
		return false;
	}
#else
	/* SoundAtomCue assetをシーケンサー上に配置したかどうかの判定 */
	if (!Track->IsA(UMovieSceneAtomTrack::StaticClass())) {
		return false;
	}
#endif

	TSharedPtr<FDragDropOperation> Operation = DragDropEvent.GetOperation();

	if (!Operation.IsValid() || !Operation->IsOfType<FAssetDragDropOp>()) {
		return false;
	}

	TSharedPtr<FAssetDragDropOp> DragDropOp = StaticCastSharedPtr<FAssetDragDropOp>(Operation);

	for (const FAssetData& AssetData : DragDropOp->GetAssets()) {
		if (Cast<USoundAtomCue>(AssetData.GetAsset())) {
			return true;
		}
	}

	return false;
}

#if ENGINE_MAJOR_VERSION == 4
FReply FAtomTrackEditor::OnDrop(const FDragDropEvent& DragDropEvent, UMovieSceneTrack* Track, int32 RowIndex, const FGuid& TargetObjectGuid)
#else
FReply FAtomTrackEditor::OnDrop(const FDragDropEvent& DragDropEvent, const FSequencerDragDropParams& DragDropParams)
#endif
{
#if ENGINE_MAJOR_VERSION == 5
	if (!DragDropParams.Track.IsValid() || !DragDropParams.Track->IsA(UMovieSceneAtomTrack::StaticClass())) {
		return FReply::Unhandled();
	}
#else
	/* SoundAtomCue assetをシーケンサー上に配置したかどうかの判定 */
	if (!Track->IsA(UMovieSceneAtomTrack::StaticClass())) {
		return FReply::Unhandled();
	}
#endif
	TSharedPtr<FDragDropOperation> Operation = DragDropEvent.GetOperation();

	if (!Operation.IsValid() || !Operation->IsOfType<FAssetDragDropOp>()) {
		return FReply::Unhandled();
	}

	TSharedPtr<FAssetDragDropOp> DragDropOp = StaticCastSharedPtr<FAssetDragDropOp>(Operation);

	bool bAnyDropped = false;
	for (const FAssetData& AssetData : DragDropOp->GetAssets()) {
		USoundAtomCue* Sound = Cast<USoundAtomCue>(AssetData.GetAsset());

		if (Sound) {
			/* SoundAtomCueがドロップされたらマスターサウンドセクションを追加 */
#if ENGINE_MAJOR_VERSION == 5
			int32 RowIndex = DragDropParams.RowIndex;
#endif
			AnimatablePropertyChanged(FOnKeyProperty::CreateRaw(this, &FAtomTrackEditor::AddNewMasterSound, Sound, RowIndex));

			bAnyDropped = true;
		}
	}

	return bAnyDropped ? FReply::Handled() : FReply::Unhandled();
}

TSharedRef<ISequencerSection> FAtomTrackEditor::MakeSectionInterface(UMovieSceneSection& SectionObject, UMovieSceneTrack& Track, FGuid ObjectBinding)
{
	check(SupportsType(SectionObject.GetOuter()->GetClass()));
	return MakeShareable(new FAtomSection(SectionObject, GetSequencer()));
}

TSharedPtr<SWidget> FAtomTrackEditor::BuildOutlinerEditWidget(const FGuid& ObjectBinding, UMovieSceneTrack* Track, const FBuildEditWidgetParams& Params)
{
	// Create a container edit box
	return SNew(SHorizontalBox)

	// Add the audio combo box
	+ SHorizontalBox::Slot()
	.AutoWidth()
	.VAlign(VAlign_Center)
	[
		FSequencerUtilities::MakeAddButton(LOCTEXT("AtomText", "Atom"), FOnGetContent::CreateSP(this, &FAtomTrackEditor::BuildAtomSubMenu, FOnAssetSelected::CreateRaw(this, &FAtomTrackEditor::OnAtomAssetSelected, Track), FOnAssetEnterPressed::CreateRaw(this, &FAtomTrackEditor::OnAtomAssetEnterPressed, Track)), Params.NodeIsHovered, GetSequencer())
	];
}

bool FAtomTrackEditor::HandleAssetAdded(UObject* Asset, const FGuid& TargetObjectGuid)
{
	if (Asset->IsA<USoundAtomCue>()) {
		auto Sound = Cast<USoundAtomCue>(Asset);

		const FScopedTransaction Transaction(LOCTEXT("AddAtom_Transaction", "Add Atom"));

		if (TargetObjectGuid.IsValid()) {
			TArray<TWeakObjectPtr<>> OutObjects;
			for (TWeakObjectPtr<> Object : GetSequencer()->FindObjectsInCurrentSequence(TargetObjectGuid)) {
				OutObjects.Add(Object);
			}
			/* Guid指定されたオブジェクトがある時はオブジェクトに紐づく形でAtomTrackを追加 */
			AnimatablePropertyChanged(FOnKeyProperty::CreateRaw(this, &FAtomTrackEditor::AddNewAttachedSound, Sound, OutObjects));
		} else {
			/* Guid指定のオブジェクトがない場合はマスタートラックとしてAtomTrackを追加*/
			int32 RowIndex = INDEX_NONE;
			AnimatablePropertyChanged( FOnKeyProperty::CreateRaw(this, &FAtomTrackEditor::AddNewMasterSound, Sound, RowIndex));
		}

		return true;
	}
	return false;
}

FKeyPropertyResult FAtomTrackEditor::AddNewMasterSound(FFrameNumber KeyTime, USoundAtomCue* Sound, int32 RowIndex)
{
	FKeyPropertyResult KeyPropertyResult;

	UMovieScene* FocusedMovieScene = GetFocusedMovieScene();
	FocusedMovieScene->Modify();

	FFindOrCreateMasterTrackResult<UMovieSceneAtomTrack> TrackResult = FindOrCreateMasterTrack<UMovieSceneAtomTrack>();
	UMovieSceneTrack* Track = TrackResult.Track;
	Track->Modify();

	auto AtomTrack = Cast<UMovieSceneAtomTrack>(Track);
	UMovieSceneSection* NewSection = AtomTrack->AddNewSoundOnRow(Sound, KeyTime, RowIndex);

	if (TrackResult.bWasCreated) {
		AtomTrack->SetDisplayName(LOCTEXT("AtomTrackName", "Atom"));
	}

	KeyPropertyResult.bTrackModified = true;

	GetSequencer()->EmptySelection();
	GetSequencer()->SelectSection(NewSection);
	GetSequencer()->ThrobSectionSelection();

	return KeyPropertyResult;
}

FKeyPropertyResult FAtomTrackEditor::AddNewAttachedSound(FFrameNumber KeyTime, USoundAtomCue* Sound, TArray<TWeakObjectPtr<UObject>> ObjectsToAttachTo)
{
	FKeyPropertyResult KeyPropertyResult;

	for(int32 ObjectIndex = 0; ObjectIndex < ObjectsToAttachTo.Num(); ++ObjectIndex) {
		UObject* Object = ObjectsToAttachTo[ObjectIndex].Get();

		FFindOrCreateHandleResult HandleResult = FindOrCreateHandleToObject(Object);
		FGuid ObjectHandle = HandleResult.Handle;
		KeyPropertyResult.bHandleCreated |= HandleResult.bWasCreated;

		if (ObjectHandle.IsValid()) {
			FFindOrCreateTrackResult TrackResult = FindOrCreateTrackForObject(ObjectHandle, UMovieSceneAtomTrack::StaticClass());
			UMovieSceneTrack* Track = TrackResult.Track;
			KeyPropertyResult.bTrackCreated |= TrackResult.bWasCreated;

			if (ensure(Track)) {
				auto AtomTrack = Cast<UMovieSceneAtomTrack>(Track);
				AtomTrack->Modify();
				UMovieSceneSection* NewSection = AtomTrack->AddNewSound(Sound, KeyTime);
				AtomTrack->SetDisplayName(LOCTEXT("AtomTrackName", "Atom"));
				KeyPropertyResult.bTrackModified = true;

				GetSequencer()->EmptySelection();
				GetSequencer()->SelectSection(NewSection);
				GetSequencer()->ThrobSectionSelection();
			}
		}
	}

	return KeyPropertyResult;
}

void FAtomTrackEditor::HandleAddAtomTrackMenuEntryExecute()
{
	/* AtomTrackの作成処理（Sectionは作成しない） */
	UMovieScene* FocusedMovieScene = GetFocusedMovieScene();

	if (FocusedMovieScene == nullptr) {
		return;
	}

	const FScopedTransaction Transaction(NSLOCTEXT("Sequencer", "AddAtomTrack_Transaction", "Add Atom Track"));
	FocusedMovieScene->Modify();

	auto NewTrack = FocusedMovieScene->AddMasterTrack<UMovieSceneAtomTrack>();
	ensure(NewTrack);
	NewTrack->SetDisplayName(LOCTEXT("AtomTrackName", "Atom"));

	if (GetSequencer().IsValid()) {
#if (ENGINE_MAJOR_VERSION == 4 && ENGINE_MINOR_VERSION > 22) || ENGINE_MAJOR_VERSION == 5
		GetSequencer()->OnAddTrack(NewTrack, FGuid());
#else
		GetSequencer()->OnAddTrack(NewTrack);
#endif
	}

	GetSequencer()->NotifyMovieSceneDataChanged(EMovieSceneDataChangeType::MovieSceneStructureItemAdded);
}

void FAtomTrackEditor::HandleAddAttachedAtomTrackMenuEntryExecute(FMenuBuilder& MenuBuilder, TArray<FGuid> ObjectBindings)
{
	MenuBuilder.AddWidget(BuildAtomSubMenu(FOnAssetSelected::CreateRaw(this, &FAtomTrackEditor::OnAttachedAtomAssetSelected, ObjectBindings), FOnAssetEnterPressed::CreateRaw(this, &FAtomTrackEditor::OnAttachedAtomEnterPressed, ObjectBindings)), FText::GetEmpty(), true);
}

TSharedRef<SWidget> FAtomTrackEditor::BuildAtomSubMenu(FOnAssetSelected OnAssetSelected, FOnAssetEnterPressed OnAssetEnterPressed)
{
	/* トラックからサブトラックとしてAtomTrackを追加する際のメニュー表示用処理 */
	FAssetRegistryModule& AssetRegistryModule = FModuleManager::LoadModuleChecked<FAssetRegistryModule>(TEXT("AssetRegistry"));

	
#if ENGINE_MAJOR_VERSION == 5 && ENGINE_MINOR_VERSION > 0
	TArray<FTopLevelAssetPath> ClassNames;
	ClassNames.Add(USoundAtomCue::StaticClass()->GetClassPathName());
	TSet<FTopLevelAssetPath> DerivedClassNames;
	AssetRegistryModule.Get().GetDerivedClassNames(ClassNames, TSet<FTopLevelAssetPath>(), DerivedClassNames);
#else
	TArray<FName> ClassNames;
	ClassNames.Add(USoundAtomCue::StaticClass()->GetFName());
	TSet<FName> DerivedClassNames;
	AssetRegistryModule.Get().GetDerivedClassNames(ClassNames, TSet<FName>(), DerivedClassNames);
#endif

	FMenuBuilder MenuBuilder(true, nullptr);

	FAssetPickerConfig AssetPickerConfig;
	{
		AssetPickerConfig.OnAssetSelected = OnAssetSelected;
		AssetPickerConfig.OnAssetEnterPressed = OnAssetEnterPressed;
		AssetPickerConfig.bAllowNullSelection = false;
		AssetPickerConfig.InitialAssetViewType = EAssetViewType::List;
		for (auto ClassName : DerivedClassNames) {
#if ENGINE_MAJOR_VERSION == 5 && ENGINE_MINOR_VERSION > 0
			AssetPickerConfig.Filter.ClassPaths.Add(ClassName);
#else
			AssetPickerConfig.Filter.ClassNames.Add(ClassName);
#endif
		}
	}

	FContentBrowserModule& ContentBrowserModule = FModuleManager::Get().LoadModuleChecked<FContentBrowserModule>(TEXT("ContentBrowser"));

	/* コンテンツブラウザ上にあるSoundAtomCueをコンボボックスにまとめる */
	TSharedPtr<SBox> MenuEntry = SNew(SBox)
		.WidthOverride(300.0f)
		.HeightOverride(300.f)
		[
			ContentBrowserModule.Get().CreateAssetPicker(AssetPickerConfig)
		];

	MenuBuilder.AddWidget(MenuEntry.ToSharedRef(), FText::GetEmpty(), true);

	return MenuBuilder.MakeWidget();
}

void FAtomTrackEditor::OnAtomAssetSelected(const FAssetData& AssetData, UMovieSceneTrack* Track)
{
	/* SoundCueAssetが選択された際の処理（MasterTrackへの追加でSoundAtomCueアセットを選んだ時などに呼ばれる） */
	FSlateApplication::Get().DismissAllMenus();

	UObject* SelectedObject = AssetData.GetAsset();

	if (SelectedObject) {
		USoundAtomCue* NewSound = CastChecked<USoundAtomCue>(AssetData.GetAsset());
		if (NewSound != nullptr) {
			const FScopedTransaction Transaction(NSLOCTEXT("Sequencer", "AddAtom_Transaction", "Add Atom"));

			auto AtomTrack = Cast<UMovieSceneAtomTrack>(Track);
			AtomTrack->Modify();

			FFrameTime KeyTime = GetSequencer()->GetLocalTime().Time;
			UMovieSceneSection* NewSection = AtomTrack->AddNewSound(NewSound, KeyTime.FrameNumber);

			GetSequencer()->EmptySelection();
			GetSequencer()->SelectSection(NewSection);
			GetSequencer()->ThrobSectionSelection();

			GetSequencer()->NotifyMovieSceneDataChanged(EMovieSceneDataChangeType::MovieSceneStructureItemAdded);
		}
	}
}

void FAtomTrackEditor::OnAtomAssetEnterPressed(const TArray<FAssetData>& AssetData, UMovieSceneTrack* Track)
{
	/* Asset選択時にエンターキーで決定した際の処理 */
	if (AssetData.Num() > 0) {
		OnAtomAssetSelected(AssetData[0].GetAsset(), Track);
	}
}

void FAtomTrackEditor::OnAttachedAtomAssetSelected(const FAssetData& AssetData, TArray<FGuid> ObjectBindings)
{
	/* Actor等のオブジェクトの下にAtomTrackをアタッチしようとした際に呼ばれる */
	FSlateApplication::Get().DismissAllMenus();

	UObject* SelectedObject = AssetData.GetAsset();

	if (SelectedObject)
	{
		const FScopedTransaction Transaction(NSLOCTEXT("Sequencer", "AddAudio_Transaction", "Add Audio"));

		for (FGuid ObjectBinding : ObjectBindings)
		{
			HandleAssetAdded(SelectedObject, ObjectBinding);
		}
	}
}

void FAtomTrackEditor::OnAttachedAtomEnterPressed(const TArray<FAssetData>& AssetData, TArray<FGuid> ObjectBindings)
{
	/* Asset選択時にエンターキーで決定した際の処理 */
	if (AssetData.Num() > 0)
	{
		OnAttachedAtomAssetSelected(AssetData[0].GetAsset(), ObjectBindings);
	}
}

FAtomSection::FAtomSection(UMovieSceneSection& InSection, TWeakPtr<ISequencer> InSequencer)
	: Sequencer(InSequencer)
	, Section(InSection)
	, StoredDrawRange(TRange<float>::Empty())
	, StoredSectionHeight(0.f)
	, InitialStartOffsetDuringResize(0)
	, InitialStartTimeDuringResize(0)
{
}

FAtomSection::~FAtomSection()
{
}

UMovieSceneSection* FAtomSection::GetSectionObject()
{
	return &Section;
}

FText FAtomSection::GetSectionTitle() const
{
	UMovieSceneAtomSection* AtomSection = Cast<UMovieSceneAtomSection>(&Section);
	if (AtomSection && AtomSection->GetSound()) {
		return FText::FromString(AtomSection->GetSound()->GetName());
	}

	return NSLOCTEXT("FAtomSection", "NoAtomTitleName", "No Atom");
}

float FAtomSection::GetSectionHeight() const
{
	return Section.GetTypedOuter<UMovieSceneAtomTrack>()->GetRowHeight();
}

int32 FAtomSection::OnPaintSection(FSequencerSectionPainter& Painter) const
{
	int32 LayerId = Painter.PaintSectionBackground();

	if (WaveformThumbnail.IsValid() && WaveformThumbnail->ShouldRender())
	{
		auto TintColor = FLinearColor::White;

		USoundAtomCue* SoundAtomCue = StoredSoundAtomCue.Get();
		if (SoundAtomCue)
		{
			TintColor = SoundAtomCue->CueColor;
		}

		// @todo Sequencer draw multiple times if looping possibly - requires some thought about SoundCues
		FSlateDrawElement::MakeViewport(
			Painter.DrawElements,
			++LayerId,
			Painter.SectionGeometry.ToPaintGeometry(FVector2D(StoredXOffset, 0), FVector2D(StoredXSize, GetSectionHeight() + 8.f)),
			WaveformThumbnail,
			(Painter.bParentEnabled ? ESlateDrawEffect::None : ESlateDrawEffect::DisabledEffect) | ESlateDrawEffect::NoGamma,
			TintColor
		);
	}


	/* 波形の終端を示す縦線を描画 */

	const ESlateDrawEffect DrawEffects = Painter.bParentEnabled ? ESlateDrawEffect::None : ESlateDrawEffect::DisabledEffect;

	const FTimeToPixel& TimeToPixelConverter = Painter.GetTimeConverter();

	static const FSlateBrush* GenericDivider = FCoreStyle::Get().GetBrush("GenericWhiteBox");

	if (!Section.HasStartFrame() || !Section.HasEndFrame())
	{
		return LayerId;
	}

	UMovieSceneAtomSection* AudioSection = Cast<UMovieSceneAtomSection>(&Section);
	if (!AudioSection || !AudioSection->GetSound())
	{
		return LayerId;
	}

	FFrameRate TickResolution = TimeToPixelConverter.GetTickResolution();
	const float AudioDuration = AudioSection->GetSound()->GetFirstWaveDuration();

	if (AudioDuration > KINDA_SMALL_NUMBER)
	{
		const float MaxOffset        = Section.GetRange().Size<FFrameTime>() / TickResolution;
		const float StartOffsetTime  = TickResolution.AsSeconds(AudioSection->GetStartOffset());
		const float SectionStartTime = TickResolution.AsSeconds(AudioSection->GetInclusiveStartFrame());

		float OffsetTime = AudioDuration - StartOffsetTime;
		while (OffsetTime < MaxOffset)
		{
			float OffsetPixel = TimeToPixelConverter.SecondsToPixel(SectionStartTime + OffsetTime);

			FSlateDrawElement::MakeBox(
				Painter.DrawElements,
				LayerId,
				Painter.SectionGeometry.MakeChild(
					FVector2D(2.f, Painter.SectionGeometry.Size.Y - 2.f),
					FSlateLayoutTransform(FVector2D(OffsetPixel, 1.f))
				).ToPaintGeometry(),
				GenericDivider,
				DrawEffects
			);

			OffsetTime += AudioDuration;

			if (!AudioSection->IsLooping())
			{
				break;
			}
		}
	}

	return LayerId;
}

void FAtomSection::Tick(const FGeometry& AllottedGeometry, const FGeometry& ParentGeometry, const double InCurrentTime, const float InDeltaTime)
{
	UMovieSceneAtomSection* AtomSection = Cast<UMovieSceneAtomSection>(&Section);
	UMovieSceneTrack* Track = Section.GetTypedOuter<UMovieSceneTrack>();

	USoundAtomCue* SoundAtomCue = AtomSection->GetSound();
	if (Track && SoundAtomCue) {
		const FSlateRect ParentRect = TransformRect(
			Concatenate(
				ParentGeometry.GetAccumulatedLayoutTransform(),
				AllottedGeometry.GetAccumulatedLayoutTransform().Inverse()
			),
			FSlateRect(FVector2D(0, 0), ParentGeometry.GetLocalSize())
		);

		const float LeftMostVisiblePixel = FMath::Max(ParentRect.Left, 0.f);
		const float RightMostVisiblePixel = FMath::Min(ParentRect.Right, AllottedGeometry.GetLocalSize().X);

		FFrameRate   TickResolution = AtomSection->GetTypedOuter<UMovieScene>()->GetTickResolution();
		FTimeToPixel TimeToPixel( AllottedGeometry, AtomSection->GetRange() / TickResolution, TickResolution );

		TRange<float> DrawRange = TRange<float>(
			TimeToPixel.PixelToSeconds(LeftMostVisiblePixel),
			TimeToPixel.PixelToSeconds(RightMostVisiblePixel)
			);

		// generate texture x offset and x size
		int32 XOffset = LeftMostVisiblePixel;//PixelRange.GetLowerBoundValue() - TimeToPixelConverter.TimeToPixel(SectionRange.GetLowerBoundValue());
		int32 XSize = RightMostVisiblePixel - LeftMostVisiblePixel;//PixelRange.Size<int32>();

		if (!FMath::IsNearlyEqual(DrawRange.GetLowerBoundValue(), StoredDrawRange.GetLowerBoundValue()) ||
			!FMath::IsNearlyEqual(DrawRange.GetUpperBoundValue(), StoredDrawRange.GetUpperBoundValue()) ||
			XOffset != StoredXOffset || XSize != StoredXSize || Track->GetColorTint() != StoredColor ||
			StoredSoundAtomCue != SoundAtomCue ||
			StoredSectionHeight != GetSectionHeight() ||
			StoredStartOffset != AtomSection->GetStartOffset() || 
			StoredLoop != AtomSection->IsLooping())
		{
			float DisplayScale = XSize / DrawRange.Size<float>();

			// Use the view range if possible, as it's much more stable than using the texture size and draw range
			TSharedPtr<ISequencer> SequencerPin = Sequencer.Pin();
			if (SequencerPin.IsValid())
			{
				DisplayScale = SequencerPin->GetViewRange().Size<float>() / ParentGeometry.GetLocalSize().X;
			}

			RegenerateWaveforms(DrawRange, XOffset, XSize, Track->GetColorTint(), DisplayScale);
			StoredSoundAtomCue = SoundAtomCue;
		}
	} else {
		/* トラック、SoundAtomCueがない場合はWabeformThumbnailの情報をリセットする */
		WaveformThumbnail.Reset();
		StoredDrawRange = TRange<float>::Empty();
		StoredSoundAtomCue.Reset();
		return;
	}
}

void FAtomSection::BeginSlipSection()
{
	UMovieSceneAtomSection* AtomSection = Cast<UMovieSceneAtomSection>(&Section);
	InitialStartOffsetDuringResize = AtomSection->GetStartOffset();
	double StartFrame = static_cast<double>(AtomSection->HasStartFrame() ? AtomSection->GetInclusiveStartFrame().Value : 0.0);
	double TickResolution = AtomSection->GetTypedOuter<UMovieScene>()->GetTickResolution().AsDecimal();
	InitialStartTimeDuringResize   = FFrameNumber(static_cast<int32>(StartFrame / TickResolution));
}

void FAtomSection::SlipSection(FFrameNumber SlipTime)
{
	UMovieSceneAtomSection* AtomSection = Cast<UMovieSceneAtomSection>(&Section);

	FFrameNumber StartOffset = (SlipTime - InitialStartTimeDuringResize);
	StartOffset += InitialStartOffsetDuringResize;

	// Ensure start offset is not less than 0
	StartOffset = FMath::Max(StartOffset.Value, 0);

	AtomSection->SetStartOffset(StartOffset.Value);

	ISequencerSection::SlipSection(SlipTime);
}

void FAtomSection::RegenerateWaveforms(TRange<float> DrawRange, int32 XOffset, int32 XSize, const FColor& ColorTint, float DisplayScale)
{
	UMovieSceneAtomSection* AtomSection = Cast<UMovieSceneAtomSection>(&Section);

	StoredDrawRange = DrawRange;
	StoredXOffset = XOffset;
	StoredXSize = XSize;
	StoredColor = ColorTint;
	StoredStartOffset = AtomSection->GetStartOffset();
	StoredSectionHeight = GetSectionHeight();
	StoredLoop = AtomSection->IsLooping();

	if (DrawRange.IsDegenerate() || DrawRange.IsEmpty() || AtomSection->GetSound() == NULL) {
		WaveformThumbnail.Reset();
	} else {
		WaveformThumbnail = MakeShareable(new FAtomThumbnail(Section, DrawRange, XSize, ColorTint, DisplayScale));
	}
}

#undef LOCTEXT_NAMESPACE
